/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.algo;

import java.io.PrintStream;

import at.jku.risc.stout.urauc.algo.AntiUnifyProblem.CommutativeArrangementIter;
import at.jku.risc.stout.urauc.algo.Contour.ContourAtom;
import at.jku.risc.stout.urauc.data.Alignment;
import at.jku.risc.stout.urauc.data.TermAtomList;
import at.jku.risc.stout.urauc.data.atom.TermAtom;
import at.jku.risc.stout.urauc.util.IntList;

/**
 * Implementation of an alignment computation function with longest subsequence.<br>
 * This implementation combines the common subsequence generator and the
 * admissibility test. It generates all the admissible alignments of two given
 * sequences of a certain length.
 * 
 * @see AlignFnc
 * @author Alexander Baumgartner
 */
public class AlignFncLAA extends AlignFnc {
	// List of contours
	private ContourArrangement allContours = new ContourArrangement();
	private ContourArrangement someContours;
	// Iterator state
	private Contour[] iterClipArray;
	private int[] iterClipIdx;
	private int iterArrayIdx;
	// Reused alignment
	private Alignment alignmentRet = new Alignment();
	private int alignLen;
	private boolean nextReady;

	@Override
	public AlignmentIterator getIterator(AntiUnifyProblem aup, int minAlignLen,
			int upperBound, boolean iterateAll, PrintStream out)
			throws MaximumIterationException {
		// Determine arrangements with maximum length.
		IntList maxLenIdx = new IntList(128);

		int len = Math.max(minAlignLen, 1);
		CommutativeArrangementIter i = aup.iteratorHorizontalPartX(true);
		long idx = 0;
		if (len <= i.getMaxPossible()) {
			while (i.hasNext()) {
				if (idx == upperBound && idx > 0)
					throw new MaximumIterationException("Upper bound of "
							+ upperBound + " iterated arrangements reached!",
							createAlignmentIterator(aup, maxLenIdx, len));
				idx++;
				i.next();
				int alignLen = initArrangement(i.getLeftWord(),
						i.getRightWord(), len, i.getMaxPossible());
				if (debugLevel != DebugLevel.SILENT && idx % 1000000 == 0)
					out.println(idx / 1000000
							+ " million commutative arrangements tested."
							+ " (Max. alignment length = " + len + ")");
				if (alignLen >= len) {
					if (alignLen > len) {
						len = alignLen;
						maxLenIdx.size = 0;
						maxLenIdx.add(idx - 1);
						if (len == i.getMaxPossible() && !iterateAll)
							break; // fond one with max. possible length
						if (len == i.getLeftWord().size
								&& len == i.getRightWord().size)
							break; // equal input

					} else {
						maxLenIdx.add(idx - 1);
					}
					if (debugLevel == DebugLevel.PROGRESS
							&& maxLenIdx.size % 1000 == 0)
						out.println(maxLenIdx.size
								+ " commutative arrangements"
								+ " with alignment length " + alignLen
								+ " found.");
				}
			}
		}

		if (debugLevel == DebugLevel.PROGRESS
				|| debugLevel == DebugLevel.VERBOSE) {
			out.println(idx + " commutative arrangements tested.");
			out.println(maxLenIdx.size + " commutative arrangements"
					+ " with alignment length " + alignLen + " found.");
		}

		return createAlignmentIterator(aup, maxLenIdx, len);
	}

	private AlignmentIterator createAlignmentIterator(AntiUnifyProblem aup,
			IntList maxLenIdx, int len) {
		if (maxLenIdx.size == 0) {
			AlignFncInput ret = new AlignFncInput();
			ret.setAlignment(new Alignment());
			return ret;
		}
		return new AlignmentIter(aup.iteratorHorizontalPartX(true), maxLenIdx,
				len);
	}

	private int initArrangement(TermAtomList left, TermAtomList right,
			int minAlignLen, int maxAlignLen) {
		if (minLenFastTest(left, right, minAlignLen))
			return -1;

		initAllContours(left, right);
		if (debugLevel == null)
			System.out.println(allContours); // Debug to stdout instead of out
		if (allContours.size() < maxAlignLen)
			maxAlignLen = allContours.size();
		if (maxAlignLen < minAlignLen)
			return -1;
		iterClipArray = new Contour[maxAlignLen];
		iterClipIdx = new int[maxAlignLen + 1];

		if (!existsAlignmentOfLen(minAlignLen))
			return -1;
		if (maxAlignLen == minAlignLen)
			return minAlignLen;
		minAlignLen++;

		while (true) { // Weighted bisection which tends to the lower bound
			int midLen = (3 * minAlignLen + maxAlignLen) / 4;
			// int midLen = (minAlignLen + maxAlignLen) / 2;
			if (existsAlignmentOfLen(midLen)) {
				if (midLen == maxAlignLen)
					return midLen;
				minAlignLen = midLen + 1;
			} else {
				if (midLen == minAlignLen)
					return minAlignLen - 1;
				maxAlignLen = midLen;
			}
		}
	}

	// Fast and stable over-approximation of the alignment length
	private boolean minLenFastTest(TermAtomList left, TermAtomList right,
			int minLen) {
		if (minLen < 3) // fast test does not pay off
			return false;
		int lenJ = right.size;
		String[] rightNames = new String[lenJ];
		for (int j = lenJ; --j >= 0;)
			rightNames[j] = right.atoms[j].getName();
		int[] prevLenI = new int[lenJ + 1];
		for (int curJ1, i = left.size; --i >= 0; prevLenI[0] = curJ1) {
			String leftA = left.atoms[i].getName();
			curJ1 = 0;
			for (int curJ0 = 0, j = lenJ; --j >= 0; curJ1 = curJ0) {
				if (leftA == rightNames[j]) {
					curJ0 = prevLenI[j + 1] + 1;
					if (curJ0 == minLen)
						return false;
				} else if (prevLenI[j] > curJ1) {
					curJ0 = prevLenI[j];
				}
				prevLenI[j + 1] = curJ1;
			}
		}
		return true;
	}

	private boolean existsAlignmentOfLen(int alignLen) {
		this.alignLen = alignLen;
		nextReady = false;
		alignmentRet.setSize(alignLen);
		allContours.initIterator(alignLen);
		iterArrayIdx = alignLen;
		return hasNext();
	}

	// Trace the common subsequence contours
	private void initAllContours(TermAtomList left, TermAtomList right) {
		allContours.clear();
		int lenJ = right.size;
		int[] prevLenI = new int[lenJ + 1];
		for (int i = 0; i < left.size; i++) {
			int prevLen = 0;
			TermAtom tLeft = left.atoms[i];
			for (int j = 0, currLen = 0; j < lenJ; prevLen = currLen) {
				if (tLeft.equals(right.atoms[j])) {
					currLen = prevLenI[j] + 1;
					allContours.setMatchAtom(currLen - 1, i, j, tLeft,
							right.atoms[j]);
				} else if (prevLenI[j + 1] > prevLen) {
					currLen = prevLenI[j + 1];
				}
				prevLenI[j] = prevLen;
				++j;
			}
			prevLenI[lenJ] = prevLen;
		}
	}

	private boolean hasNext() {
		if (nextReady)
			return true;
		while (true) {
			if (iterArrayIdx >= alignLen && !nextContourArrangement())
				return nextReady = false;
			// If the alignment length is 0 then there are no contours
			if (alignLen == 0) {
				iterArrayIdx = 0;
				return nextReady = true;
			}
			// Walk down from contour N to contour 0 to obtain the next
			// alignment
			Contour cl = iterClipArray[iterArrayIdx];
			int iterClipStart = iterClipIdx[iterArrayIdx];
			while (nextAdmissibleAtom(cl, iterClipStart, iterArrayIdx)) {
				iterClipArray[iterArrayIdx] = cl;
				if (iterArrayIdx == 0)
					return nextReady = true;
				ContourAtom el = cl.get(iterClipIdx[iterArrayIdx]);
				iterArrayIdx--;
				cl = someContours.getClip(iterArrayIdx, el);
				iterClipStart = 0;
			}
			iterArrayIdx++;
			nextContourBranch();
		}
	}

	private boolean nextAdmissibleAtom(Contour cl, int i, int idx) {
		for (int n = cl.size(); i < n; i++) {
			if (alignmentRet.setAtomIfTailAdmissible(idx, cl.get(i).atom)) {
				iterClipIdx[idx] = i;
				return true;
			}
		}
		return false;
	}

	private void nextContourBranch() {
		// Walk up the contours until we reach a new branch
		while (iterArrayIdx < alignLen
				&& iterClipIdx[iterArrayIdx] + 1 >= iterClipArray[iterArrayIdx]
						.size())
			iterArrayIdx++;
		iterClipIdx[iterArrayIdx]++; // no need to test idx < alignLen
	}

	private boolean nextContourArrangement() {
		if (allContours.hasNext()) {
			someContours = allContours.next();
			int idx = alignLen - 1;
			iterArrayIdx = idx;
			// If the alignment length is 0 then there are no contours
			if (idx >= 0) {
				iterClipArray[idx] = someContours.getContour(idx);
				iterClipIdx[idx] = 0;
			}
			return true;
		}
		return false;
	}

	/**
	 * ATTENTION: This method always returns the same alignment object! You have
	 * to copy the alignment if you still need it after the next iteration step
	 * ({@linkplain #hasNext()} will overwrite it).
	 */
	private Alignment next() {
		nextContourBranch();
		nextReady = false;
		return alignmentRet;
	}

	@Override
	public AlignFnc newInstance() {
		return new AlignFncLAA();
	}

	public static class AlignmentIter implements AlignmentIterator {
		private AlignFncLAA alignFnc = new AlignFncLAA();
		private CommutativeArrangementIter iter;
		private IntList consideredArrangement;
		private int alignLen;
		private int idx;
		private long count;

		public AlignmentIter(CommutativeArrangementIter iter,
				IntList consideredArrangement, int alignLen) {
			this.iter = iter;
			this.consideredArrangement = consideredArrangement;
			this.alignLen = alignLen;
		}

		@Override
		public boolean hasNext() {
			return idx < consideredArrangement.size || alignFnc.hasNext();
		}

		@Override
		public Alignment next() {
			if (alignFnc.hasNext())
				return alignFnc.next();
			for (long nextArrangement = consideredArrangement.values[idx]; count <= nextArrangement; count++)
				iter.next();
			idx++;
			if (alignFnc.initArrangement(iter.getLeftWord(),
					iter.getRightWord(), alignLen, alignLen) == -1)
				throw new IllegalStateException(
						"Illegal arrangement for alignment length " + alignLen);
			return next();
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException("Remove is not supported.");
		}

		@Override
		public int getCommutativeArrangementSize() {
			return consideredArrangement.size;
		}

		@Override
		public int getMaxAlignmentLen() {
			return alignLen;
		}
	}
}
