/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.algo;

import java.io.IOException;
import java.io.StringWriter;
import java.util.Collections;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import at.jku.risc.stout.urauc.data.Alignment;
import at.jku.risc.stout.urauc.data.Equation;
import at.jku.risc.stout.urauc.data.Hedge;
import at.jku.risc.stout.urauc.data.NodeFactory;
import at.jku.risc.stout.urauc.data.TermAtomList;
import at.jku.risc.stout.urauc.data.TermNode;
import at.jku.risc.stout.urauc.data.atom.ContextVar;
import at.jku.risc.stout.urauc.data.atom.HedgeVar;
import at.jku.risc.stout.urauc.data.atom.TermAtom;
import at.jku.risc.stout.urauc.data.atom.Variable;
import at.jku.risc.stout.urauc.util.DataStructureFactory;

/**
 * This class represents an anti-unification problem (AUP) which consists of one
 * generalization variable (the most general generalization), and two
 * {@linkplain Hedge}s. The hedges represent the left hand side and the right
 * hand side of the equation.
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnifyProblem implements Equation, Cloneable {
	private VariableWithHedges horizontalPartX;
	private VariableWithHedges verticalPartC;
	private Alignment alignment;

	public static String PRINT_EQ_SEPARATOR = " =^= ";
	public static String PRINT_VAR_SEPARATOR = ": ";
	public static String PRINT_SYS_SEPARATOR = "; ";

	/**
	 * Create an AUP with fresh generalization variables and the given hedges.
	 */
	public AntiUnifyProblem(Hedge left, Hedge right, Alignment alignment) {
		Variable varX = NodeFactory.obtainFreeHedgeVar();
		this.horizontalPartX = new VariableWithHedges(varX, left, right);
		Variable varC = NodeFactory.obtainCorrespondingCtxVar(varX.getName());
		Hedge ctxLeft = new Hedge(NodeFactory.obtainHole());
		ctxLeft.setCommutative(left.isCommutative());
		Hedge ctxRight = new Hedge(NodeFactory.obtainHole());
		ctxRight.setCommutative(right.isCommutative());
		this.verticalPartC = new VariableWithHedges(varC, ctxLeft, ctxRight);
		this.alignment = alignment;
	}

	private AntiUnifyProblem(VariableWithHedges horizontalPartX,
			VariableWithHedges verticalPartC, Alignment alignment) {
		this.horizontalPartX = horizontalPartX;
		this.verticalPartC = verticalPartC;
		this.alignment = alignment;
	}

	/**
	 * Obtain an empty AUP with fresh generalization variables.
	 */
	public AntiUnifyProblem() {
		this(new Hedge(), new Hedge(), null);
	}

	public void setAlignment(Alignment alignment) {
		this.alignment = alignment;
	}

	@Override
	public void addLeft(TermNode left) {
		horizontalPartX.left.add(left);
	}

	@Override
	public void addRight(TermNode right) {
		horizontalPartX.right.add(right);
	}

	/**
	 * Returns the left hedge of this AUP.
	 */
	public Hedge getLeft() {
		return horizontalPartX.left;
	}

	/**
	 * Returns the right hedge of this AUP.
	 */
	public Hedge getRight() {
		return horizontalPartX.right;
	}

	/**
	 * Creates a new hedge which is commutative iff
	 * {@linkplain #getHorizontalPartX()}.checkCommutative() is true.
	 */
	public Hedge newHedge(TermNode... nodes) {
		Hedge ret = new Hedge(horizontalPartX.checkCommutative());
		ret.addAll(nodes);
		return ret;
	}

	/**
	 * Set the left hedge of this AUP.
	 */
	public void setLeft(Hedge hedge) {
		horizontalPartX.left = hedge;
	}

	/**
	 * Set the left hedge of this AUP.
	 */
	public void setRight(Hedge hedge) {
		horizontalPartX.right = hedge;
	}

	/**
	 * Returns the horizontal abstraction of this AUP.
	 */
	public VariableWithHedges getHorizontalPartX() {
		return horizontalPartX;
	}

	/**
	 * Returns the vertical abstraction of this AUP.
	 */
	public VariableWithHedges getVerticalPartC() {
		return verticalPartC;
	}

	public Alignment getAlignment() {
		return alignment;
	}

	/**
	 * Returns the hedge generalization variable, which is used for horizontal
	 * abstraction. (It also represents the most general generalization of this
	 * AUP.)
	 */
	public HedgeVar getHorizontalVarX() {
		return (HedgeVar) horizontalPartX.var;
	}

	/**
	 * Returns the context generalization variable, which is used for vertical
	 * abstraction.
	 */
	public ContextVar getVerticalVarC() {
		return (ContextVar) verticalPartC.var;
	}

	@Override
	public String toString() {
		return horizontalPartX + PRINT_SYS_SEPARATOR + verticalPartC
				+ PRINT_SYS_SEPARATOR + alignment.toString(); // GWT needs
																// toString
	}

	/**
	 * Initialize the position indexes of the hedge (horizontal sequence
	 * {@linkplain #getHorizontalPartX()}) with the given hole position of the
	 * context ({@linkplain #getVerticalPartC()}). The child-position of the
	 * first element of the horizontal sequence is 0.
	 */
	public void initPositionIndexHorizontalPartX(int[] holeParentIndexLeft,
			int[] holeParentIndexRight) {
		initPosIdx(horizontalPartX, holeParentIndexLeft, holeParentIndexRight);
	}

	/**
	 * Initialize the position indexes of the context (vertical sequence
	 * {@linkplain #getVerticalPartC()}). The child-position of the first
	 * element of the horizontal sequence is 0.
	 */
	public void initPositionIndexVerticalPartC() {
		initPosIdx(verticalPartC, TermNode.POSITION_E, TermNode.POSITION_E);
	}

	private void initPosIdx(VariableWithHedges toInit,
			int[] holeParentIndexLeft, int[] holeParentIndexRight) {
		toInit.left.initPosition(holeParentIndexLeft);
		toInit.right.initPosition(holeParentIndexRight);
	}

	public Hedge createMostGeneral() {
		return newHedge(new TermNode(getVerticalVarC(), new Hedge(
				getHorizontalVarX())));
	}

	/**
	 * Make some symbols and/or the root hedge commutative.
	 */
	public void makeCommutativ(Set<TermAtom> commutativeSymbols,
			boolean commutativeRootHedge) {
		horizontalPartX.left.makeCommutative(commutativeSymbols,
				commutativeRootHedge);
		horizontalPartX.right.makeCommutative(commutativeSymbols,
				commutativeRootHedge);
		verticalPartC.left.makeCommutative(commutativeSymbols,
				commutativeRootHedge);
		verticalPartC.right.makeCommutative(commutativeSymbols,
				commutativeRootHedge);
	}

	/**
	 * Make some symbols and/or the root hedge commutative.
	 */
	public void makeAssociative(Set<TermAtom> associativeSymbols) {
		horizontalPartX.left.makeAssociative(associativeSymbols, null);
		horizontalPartX.right.makeAssociative(associativeSymbols, null);
		verticalPartC.left.makeAssociative(associativeSymbols, null);
		verticalPartC.right.makeAssociative(associativeSymbols, null);
	}

	public CommutativeArrangementIter iteratorHorizontalPartX(
			boolean ignoreDisjointLeaves) {
		return new CommutativeArrangementIter(horizontalPartX,
				ignoreDisjointLeaves);
	}

	public CommutativeArrangementIter iteratorVerticalPartC(
			boolean ignoreDisjointLeaves) {
		return new CommutativeArrangementIter(verticalPartC,
				ignoreDisjointLeaves);
	}

	@Override
	public AntiUnifyProblem clone() {
		return new AntiUnifyProblem(horizontalPartX.clone(),
				verticalPartC.clone(), alignment);
	}

	@Override
	public void init() {
		initPositionIndexHorizontalPartX(TermNode.POSITION_E,
				TermNode.POSITION_E);
	}

	/**
	 * @author Alexander Baumgartner
	 */
	public static class VariableWithHedges implements Cloneable {
		private Variable var;
		private Hedge left, right;

		public VariableWithHedges(Variable var, Hedge left, Hedge right) {
			this.var = var;
			this.left = left;
			this.right = right;
		}

		public Variable getVar() {
			return var;
		}

		public Hedge getLeft() {
			return left;
		}

		public Hedge getRight() {
			return right;
		}

		public void setLeft(Hedge left) {
			this.left = left;
		}

		public void setRight(Hedge right) {
			this.right = right;
		}

		@Override
		public VariableWithHedges clone() {
			return new VariableWithHedges(var, left.clone(), right.clone());
		}

		public boolean checkCommutative() {
			if (left.isCommutative() && right.isCommutative())
				return true;
			if (left.isCommutative() && right.size() < 2)
				return true;
			if (right.isCommutative() && left.size() < 2)
				return true;
			return false;
		}

		@Override
		public String toString() {
			StringWriter sb = new StringWriter();
			sb.append(var.toString());
			sb.append(PRINT_VAR_SEPARATOR);
			try {
				left.print(sb, left.size() != 1);
				sb.append(PRINT_EQ_SEPARATOR);
				right.print(sb, right.size() != 1);
			} catch (IOException e) {
			}
			return sb.toString();
		}
	}

	/**
	 * @author Alexander Baumgartner
	 */
	public class CommutativeArrangementIter {
		private boolean hasNext;
		private boolean consumed = true;
		private Hedge left;
		private Hedge right;
		private TermAtomList wordLeft;
		private TermAtomList wordRight;
		private int maxPossible;
		private boolean leftFixed;
		private boolean rightFixed;

		private CommutativeArrangementIter(VariableWithHedges iterEquation,
				boolean ignoreDisjointLeaves) {
			left = iterEquation.left;
			right = iterEquation.right;
			left.sortCommutative(true); // sort all commutative hedges
			right.sortCommutative(true); // sort all commutative hedges
			wordLeft = left.word(); // obtain word representation
			wordRight = right.word(); // obtain word representation
			Map<TermAtom, Integer> symDiff = countOcc(wordRight);
			Map<TermAtom, Integer> leftAtoms = countOcc(wordLeft);
			for (Entry<TermAtom, Integer> e : symDiff.entrySet()) {
				Integer leftCount = leftAtoms.get(e.getKey());
				if (leftCount != null)
					maxPossible += Math.min(leftCount, e.getValue());
			}
			if (ignoreDisjointLeaves) {
				Map<TermAtom, Integer> leftMinus = DataStructureFactory.$
						.newMap(leftAtoms);
				leftMinus.keySet().removeAll(symDiff.keySet());
				symDiff.keySet().removeAll(leftAtoms.keySet());
				symDiff.putAll(leftMinus); // symmetric difference
			} else {
				symDiff = Collections.emptyMap();
			}
			left.setIgnoreLeaves(symDiff);
			right.setIgnoreLeaves(symDiff);
		}

		private Map<TermAtom, Integer> countOcc(TermAtomList atomList) {
			Map<TermAtom, Integer> count = DataStructureFactory.$.newMap();
			for (int i = atomList.size; --i >= 0;) {
				Integer numOcc = count.get(atomList.atoms[i]);
				numOcc = numOcc == null ? 1 : (numOcc + 1);
				count.put(atomList.atoms[i], numOcc);
			}
			return count;
		}

		public boolean hasNext() {
			if (!consumed)
				return hasNext;
			consumed = false;
			if (!hasNext)
				return hasNext = true;
			if (!rightFixed) {
				hasNext = right.nextCommutativeArrangement();
				wordRight.size = 0;
				right.word(wordRight);
				if (hasNext)
					return true;
			}
			if (leftFixed)
				return hasNext = false;
			hasNext = left.nextCommutativeArrangement();
			wordLeft.size = 0;
			left.word(wordLeft);
			return hasNext;
		}

		public AntiUnifyProblem next() {
			if (consumed)
				hasNext();
			consumed = true;
			return AntiUnifyProblem.this;
		}

		public boolean isLeftFixed() {
			return leftFixed;
		}

		public void setLeftFixed(boolean leftFixed) {
			this.leftFixed = leftFixed;
		}

		public boolean isRightFixed() {
			return rightFixed;
		}

		public void setRightFixed(boolean rightFixed) {
			this.rightFixed = rightFixed;
		}

		public TermAtomList getLeftWord() {
			return wordLeft;
		}

		public TermAtomList getRightWord() {
			return wordRight;
		}

		public int getMaxPossible() {
			return maxPossible;
		}
	}
}
