/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.algo;

import java.io.PrintStream;
import java.util.List;

import at.jku.risc.stout.urauc.algo.AntiUnifyProblem.VariableWithHedges;
import at.jku.risc.stout.urauc.data.Alignment;
import at.jku.risc.stout.urauc.data.Alignment.AlignmentAtom;
import at.jku.risc.stout.urauc.data.EquationSystem;
import at.jku.risc.stout.urauc.data.Hedge;
import at.jku.risc.stout.urauc.data.InputParser;
import at.jku.risc.stout.urauc.data.NodeFactory;
import at.jku.risc.stout.urauc.data.TermNode;
import at.jku.risc.stout.urauc.data.atom.HedgeVar;
import at.jku.risc.stout.urauc.util.ControlledException;
import at.jku.risc.stout.urauc.util.DataStructureFactory;
import at.jku.risc.stout.urauc.util.IntList;

/**
 * This class represents a rule based system for rigid anti-unification of
 * unranked terms and hedges.<br>
 * The algorithm is described in the paper:<br>
 * <a href="http://www.risc.jku.at/home/abaumgar/">Alexander Baumgartner</a>, <a
 * href="http://www.risc.jku.at/people/tkutsia/">Temur Kutsia</a>, <a
 * href="http://drops.dagstuhl.de/opus/volltexte/2011/3118/pdf/11.pdf">
 * Anti-Unification for Unranked Terms and Hedges with Hedge- and Context
 * Variables.</a>
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnifySystem {
	private EquationSystem<AntiUnifyProblem> problemSet;
	private List<VariableWithHedges> storeVerticalQ = DataStructureFactory.$
			.newList();
	private List<VariableWithHedges> storeHorizontalS = DataStructureFactory.$
			.newList();
	private Substitution sigma;
	private boolean applyResH;
	/** TODO: Enable variable narrowing */
	protected boolean narrowVariables;

	/**
	 * Creates and initializes the rule based system.
	 */
	public AntiUnifySystem(EquationSystem<AntiUnifyProblem> problemSet,
			Substitution sigma) {
		this(problemSet, sigma, false);
	}

	/**
	 * Creates and initializes the rule based system. The experimental rule
	 * Res-H can be turned on.
	 */
	public AntiUnifySystem(EquationSystem<AntiUnifyProblem> problemSet,
			Substitution sigma, boolean applyResH) {
		this.problemSet = problemSet;
		this.sigma = sigma;
		this.applyResH = applyResH;
		for (AntiUnifyProblem aup : problemSet)
			sigma.put(aup.getHorizontalVarX(), aup.createMostGeneral());
	}

	/**
	 * Computes the result silently.
	 */
	public void compute() throws ControlledException {
		compute(DebugLevel.SILENT, null);
	}

	/**
	 * Computes the result and prints progress information to
	 * {@linkplain PrintStream} according to {@linkplain DebugLevel}.
	 */
	public void compute(DebugLevel debugLevel, PrintStream debugOut)
			throws ControlledException {
		debug(debugLevel, debugOut);
		while (!problemSet.isEmpty()) {
			AntiUnifyProblem aup = problemSet.popFirst();
			Alignment align = aup.getAlignment();
			Hedge left = aup.getLeft();
			Hedge right = aup.getRight();

			// Sol-H
			if (align.isEmpty()) {
				storeHorizontalS.add(aup.getHorizontalPartX());
				sigma.composeInRange(aup.getVerticalVarC(),
						aup.newHedge(NodeFactory.obtainHole()));
				debug("Sol-H", debugLevel, debugOut);
				continue;
			}

			AlignmentAtom alignAtom1 = align.getFirst();
			int idxLeft1 = alignAtom1.getIdxLeftTop();
			int idxRight1 = alignAtom1.getIdxRightTop();
			int depthLeft1 = alignAtom1.getIdxLeftDepth();
			int depthRight1 = alignAtom1.getIdxRightDepth();
			boolean idxLeftEqual1 = true;
			boolean idxRightEqual1 = true;
			for (int i = 1, n = align.size(); i < n && idxLeftEqual1; i++)
				idxLeftEqual1 = align.get(i).getIdxLeftTop() == idxLeft1;
			for (int i = 1, n = align.size(); i < n && idxRightEqual1; i++)
				idxRightEqual1 = align.get(i).getIdxRightTop() == idxRight1;

			// App-A
			if (idxLeftEqual1 && idxRightEqual1 && depthLeft1 == 1
					&& depthRight1 == 1) {
				// fill store
				Hedge hLeft = new Hedge(left.size(), left.isCommutative());
				hLeft.addAll(left.subHedge(0, idxLeft1));
				hLeft.add(NodeFactory.obtainHole());
				hLeft.addAll(left.subHedge(idxLeft1 + 1));
				aup.getVerticalPartC().getLeft().replaceHole(hLeft);
				Hedge hRight = new Hedge(right.size(), right.isCommutative());
				hRight.addAll(right.subHedge(0, idxRight1));
				hRight.add(NodeFactory.obtainHole());
				hRight.addAll(right.subHedge(idxRight1 + 1));
				aup.getVerticalPartC().getRight().replaceHole(hRight);
				storeVerticalQ.add(aup.getVerticalPartC());
				// decompose alignment
				align.removeFirst();
				align.incAllRootIdxLeft();
				align.incAllRootIdxRight();
				// add new AUP
				AntiUnifyProblem aupNew = new AntiUnifyProblem( //
						left.get(idxLeft1).getHedge(), //
						right.get(idxRight1).getHedge(), align);
				problemSet.addFirst(aupNew);
				// compose sigma
				Hedge hNew = aupNew.createMostGeneral();
				sigma.composeInRange(aup.getHorizontalVarX(),
						aup.newHedge(new TermNode(alignAtom1.atomLeft, hNew)));
				debug("App-A", debugLevel, debugOut);
				continue;
			}

			// Abs-L
			if (idxLeftEqual1 && depthLeft1 > 1) {
				// decompose alignment
				align.incAllRootIdxLeft();
				TermNode term = left.get(idxLeft1);
				// abstract corresponding term vertically
				aup.setLeft(term.getHedge());
				// compose vertical store
				Hedge hLeft = new Hedge(term.getHedge().isCommutative());
				hLeft.add(NodeFactory.obtainHole());
				term.setHedge(hLeft);
				aup.getVerticalPartC().getLeft().replaceHole(left);
				problemSet.addFirst(aup);
				debug("Abs-L", debugLevel, debugOut);
				continue;
			}
			// Abs-R
			if (idxRightEqual1 && depthRight1 > 1) {
				// decompose alignment
				align.incAllRootIdxRight();
				TermNode term = right.get(idxRight1);
				// abstract corresponding term vertically
				aup.setRight(term.getHedge());
				// compose vertical store
				Hedge hRight = new Hedge(term.getHedge().isCommutative());
				hRight.add(NodeFactory.obtainHole());
				term.setHedge(hRight);
				aup.getVerticalPartC().getRight().replaceHole(right);
				problemSet.addFirst(aup);
				debug("Abs-R", debugLevel, debugOut);
				continue;
			}

			// Spl-H - split AUP into two parts
			int splitAt = 1;
			for (int n = align.size(); splitAt < n; splitAt++) {
				AlignmentAtom atom = align.get(splitAt);
				if (atom.getIdxLeftTop() != idxLeft1
						&& atom.getIdxRightTop() != idxRight1)
					break;
			}

			Alignment align2 = align.subList(splitAt);
			align.setSize(splitAt);
			int idxLeftStart = align.getFirst().getIdxLeftTop();
			int idxRightStart = align.getFirst().getIdxRightTop();
			idxLeft1 = align.getLast().getIdxLeftTop() + 1;
			idxRight1 = align.getLast().getIdxRightTop() + 1;
			int idxLeftEnd = align2.getLast().getIdxLeftTop() + 1;
			int idxRightEnd = align2.getLast().getIdxRightTop() + 1;
			// fill vertical store
			Hedge ctxLeft = new Hedge(left.isCommutative());
			ctxLeft.addAll(left.subHedge(0, idxLeftStart));
			ctxLeft.add(NodeFactory.obtainHole());
			ctxLeft.addAll(left.subHedge(idxLeftEnd));
			aup.getVerticalPartC().getLeft().replaceHole(ctxLeft);
			Hedge ctxRight = new Hedge(right.isCommutative());
			ctxRight.addAll(right.subHedge(0, idxRightStart));
			ctxRight.add(NodeFactory.obtainHole());
			ctxRight.addAll(right.subHedge(idxRightEnd));
			aup.getVerticalPartC().getRight().replaceHole(ctxRight);
			storeVerticalQ.add(aup.getVerticalPartC());
			// create 2 new AUPs
			AntiUnifyProblem aup1 = new AntiUnifyProblem(left.subHedge(
					idxLeftStart, idxLeft1), right.subHedge(idxRightStart,
					idxRight1), align);
			AntiUnifyProblem aup2 = new AntiUnifyProblem(left.subHedge(
					idxLeft1, idxLeftEnd), right.subHedge(idxRight1,
					idxRightEnd), align2);
			align.addAllRootIdx(-idxLeftStart, -idxRightStart);
			align2.addAllRootIdx(-idxLeft1, -idxRight1);
			// compose sigma
			Hedge subH = aup.newHedge(aup1.createMostGeneral().get(0), aup2
					.createMostGeneral().get(0));
			sigma.composeInRange(aup.getHorizontalVarX(), subH);
			problemSet.addFirst(aup2);
			problemSet.addFirst(aup1);
			debug("Spl-H", debugLevel, debugOut);
		}

		// Res-C
		for (VariableWithHedges vertical : storeVerticalQ) {
			Hedge leftCtx = vertical.getLeft();
			Hedge rightCtx = vertical.getRight();
			if (leftCtx.size() > 1 || rightCtx.size() > 1) {
				int leftHoleIdx = leftCtx.getHoleIdx();
				int rightHoleIdx = rightCtx.getHoleIdx();
				HedgeVar x1 = NodeFactory.obtainFreeHedgeVar();
				storeHorizontalS.add(new VariableWithHedges(x1, //
						leftCtx.subHedge(0, leftHoleIdx), //
						rightCtx.subHedge(0, rightHoleIdx)));
				HedgeVar x2 = NodeFactory.obtainFreeHedgeVar();
				storeHorizontalS.add(new VariableWithHedges(x2, //
						leftCtx.subHedge(leftHoleIdx + 1), //
						rightCtx.subHedge(rightHoleIdx + 1)));
				vertical.setLeft(leftCtx.subHedge(leftHoleIdx, leftHoleIdx + 1));
				vertical.setRight(rightCtx.subHedge(rightHoleIdx,
						rightHoleIdx + 1));
				Hedge hole = new Hedge(NodeFactory.obtainHole());
				Hedge subH = new Hedge(new TermNode(x1), new TermNode(
						vertical.getVar(), hole), new TermNode(x2));
				subH.setCommutative(vertical.checkCommutative());
				sigma.composeInRange(vertical.getVar(), subH);
				debug("Res-C", debugLevel, debugOut);
			}
		}

		// Res-H
		if (applyResH)
			for (VariableWithHedges vertical : storeVerticalQ) {
				Hedge leftCtx = vertical.getLeft().getHoleHedge();
				Hedge rightCtx = vertical.getRight().getHoleHedge();
				if (leftCtx.size() > 1 || rightCtx.size() > 1) {
					int leftHoleIdx = leftCtx.getHoleIdx();
					int rightHoleIdx = rightCtx.getHoleIdx();
					HedgeVar x1 = NodeFactory.obtainFreeHedgeVar();
					storeHorizontalS.add(new VariableWithHedges(x1, //
							leftCtx.subHedge(0, leftHoleIdx), //
							rightCtx.subHedge(0, rightHoleIdx)));
					HedgeVar x2 = NodeFactory.obtainFreeHedgeVar();
					storeHorizontalS.add(new VariableWithHedges(x2, //
							leftCtx.subHedge(leftHoleIdx + 1), //
							rightCtx.subHedge(rightHoleIdx + 1)));
					Hedge hole = new Hedge(new TermNode(x1),
							NodeFactory.obtainHole(), new TermNode(x2));
					Hedge subH = new Hedge(
							new TermNode(vertical.getVar(), hole));
					subH.setCommutative((leftCtx.isCommutative() || leftCtx
							.size() == 1)
							&& (rightCtx.isCommutative() || rightCtx.size() == 1));
					leftCtx.setTo(NodeFactory.obtainHole());
					rightCtx.setTo(NodeFactory.obtainHole());
					sigma.composeInRange(vertical.getVar(), subH);
					debug("Res-H", debugLevel, debugOut);
				}
			}

		// Clr-C
		storeVerticalQ = clearStore(storeVerticalQ);
		debug("Clr-S", debugLevel, debugOut);
		// Clr-H
		storeHorizontalS = clearStore(storeHorizontalS);
		debug("Clr-S", debugLevel, debugOut);
		// Mer-C
		mergeStore("Mer-S", storeVerticalQ, debugLevel, debugOut, new Hedge(
				NodeFactory.obtainHole()));
		// Mer-H
		mergeStore("Mer-S", storeHorizontalS, debugLevel, debugOut,
				Hedge.nullHedge);
		if (narrowVariables) {
			if (narHO(storeVerticalQ, debugLevel, debugOut)) {
				mergeStore("Mer-S", storeVerticalQ, debugLevel, debugOut,
						new Hedge(NodeFactory.obtainHole()));
			}
			if (narFO(storeHorizontalS, debugLevel, debugOut)) {
				mergeStore("Mer-S", storeHorizontalS, debugLevel, debugOut,
						Hedge.nullHedge);
			}
		}
	}

	private List<VariableWithHedges> clearStore(List<VariableWithHedges> store)
			throws ControlledException {
		List<VariableWithHedges> storeClean;
		try {
			storeClean = DataStructureFactory.$.newList();
			for (VariableWithHedges eq : store) {
				if ((eq.getLeft().isHole() && eq.getRight().isHole())
						|| (eq.getLeft().isEmpty() && eq.getRight().isEmpty())) {
					sigma.composeInRange(eq.getVar(), eq.getLeft());
				} else {
					storeClean.add(eq);
				}
			}
			return storeClean;
		} catch (Exception e) {
			throw new ControlledException("Internal error in Clear Store", e) {
				private static final long serialVersionUID = 1L;
			};
		}
	}

	private void mergeStore(String ruleName, List<VariableWithHedges> store,
			DebugLevel debugLevel, PrintStream debugOut, Hedge neutral) {
		for (int i = store.size() - 2; i >= 0; i--) {
			VariableWithHedges eq1 = store.get(i);
			for (int j = store.size() - 1; j > i; j--) {
				VariableWithHedges eq2 = store.get(j);
				if (eq1.getLeft().equals(eq2.getLeft())
						&& eq1.getRight().equals(eq2.getRight())) {
					sigma.composeInRange(eq2.getVar(), new Hedge(new TermNode(
							eq1.getVar(), neutral)));
					store.remove(j);
					debug(ruleName, debugLevel, debugOut);
				}
			}
		}
	}

	/** TODO: Override and implement rule */
	protected boolean narHO(List<VariableWithHedges> storeQ,
			DebugLevel debugLevel, PrintStream debugOut) {
		boolean updated = false;
		for (int i = storeQ.size() - 1; i >= 0; i--) {
			VariableWithHedges eq = storeQ.get(i);
			IntList idxL = eq.getLeft().getHoleIdxRec();
			IntList idxR = eq.getRight().getHoleIdxRec();
			// Notice X.size > 1 for some X in {idxL, idxR} by Res-C and Clr-C
			if (idxL.size == idxR.size) {
				// Implement:
				// Decompose vertically AND CHECK that no CONTEXT variable
				// appears in the chain under the hole!!! (Attention Function
				// variables are instances of Context variables)
				// ...
				// if (conditionFulfilled) updated = true;
			}
		}
		return updated;
	}

	/** TODO: Override and implement rule */
	protected boolean narFO(List<VariableWithHedges> storeS,
			DebugLevel debugLevel, PrintStream debugOut) {
		boolean updated = false;
		for (int i = storeS.size() - 1; i >= 0; i--) {
			VariableWithHedges eq = storeS.get(i);
			// Notice eq.getX().size > 0 for some X in {Left, Right} by Clr-H
			if (eq.getLeft().size() == eq.getRight().size()) {
				// Implement:
				// Decompose horizontally AND CHECK that all elements are
				// TERMS!!! (Attention Term variables are instances of Hedge
				// variables)
				// ...
				// if (conditionFulfilled) updated = true;
			}
		}
		return updated;
	}

	private void debug(DebugLevel debugLevel, PrintStream out) {
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println(" Problem: " + problemSet);
			out.println("   Store: " + storeToString());
			out.println("   Sigma: " + sigma);
			out.println();
		}
	}

	private void debug(String rule, DebugLevel debugLevel, PrintStream out) {
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println(rule + " ==> ");
			debug(debugLevel, out);
		}
	}

	/**
	 * Returns the problem set which is empty after {@linkplain #compute()}
	 * succeeded.
	 */
	public EquationSystem<AntiUnifyProblem> getProblemSet() {
		return problemSet;
	}

	/**
	 * Returns the vertical store where all the vertical disagreements are
	 * stored during the computation. Together with the horizontal store, it
	 * contains all the information about how to instantiate the computed
	 * generalization to obtain the original terms back. (Only contains
	 * information after {@linkplain #compute()} or
	 * {@link #compute(DebugLevel, PrintStream)} succeeded.)
	 * 
	 * @see #getStoreHorizontalS()
	 */
	public List<VariableWithHedges> getStoreVerticalQ() {
		return storeVerticalQ;
	}

	/**
	 * Returns the horizontal store where all the horizontal disagreements are
	 * stored during the computation. Together with the vertical store, it
	 * contains all the information about how to instantiate the computed
	 * generalization to obtain the original terms back. (Only contains
	 * information after {@linkplain #compute()} or
	 * {@link #compute(DebugLevel, PrintStream)} succeeded.)
	 * 
	 * @see #getStoreVerticalQ()
	 */
	public List<VariableWithHedges> getStoreHorizontalS() {
		return storeHorizontalS;
	}

	/**
	 * The substitution sigma contains the generalization after
	 * {@linkplain #compute()} succeeded.
	 */
	public Substitution getSigma() {
		return sigma;
	}

	@Override
	public String toString() {
		return sigma.toRanString();
	}

	public String storeToString() {
		StringBuilder sb = new StringBuilder();
		for (VariableWithHedges varC : storeVerticalQ) {
			sb.append(", ");
			sb.append(NodeFactory.obtainHedgeVarName(varC.getVar().getName()));
			sb.append(AntiUnifyProblem.PRINT_VAR_SEPARATOR);
			sb.append("()");
			sb.append(AntiUnifyProblem.PRINT_EQ_SEPARATOR);
			sb.append("()");
			sb.append(AntiUnifyProblem.PRINT_SYS_SEPARATOR);
			sb.append(varC);
		}
		for (VariableWithHedges varX : storeHorizontalS) {
			sb.append(", ");
			sb.append(varX);
			sb.append(AntiUnifyProblem.PRINT_SYS_SEPARATOR);
			sb.append(NodeFactory.obtainCtxVarName(varX.getVar().getName()));
			sb.append(AntiUnifyProblem.PRINT_VAR_SEPARATOR);
			sb.append(InputParser.HOLE_NAME);
			sb.append(AntiUnifyProblem.PRINT_EQ_SEPARATOR);
			sb.append(InputParser.HOLE_NAME);
		}
		return sb.length() == 0 ? "" : sb.substring(2);
	}
}
