/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.algo;

import java.util.Arrays;

import at.jku.risc.stout.urauc.data.Alignment.AlignmentAtom;
import at.jku.risc.stout.urauc.data.atom.TermAtom;

/**
 * A list of contour elements. This data structure is designed to do fast filter
 * operations which are needed for subsequence determination. It uses a two
 * dimensional cache of rectangular clippings of this contour. If the caching
 * mechanism is used, then it has cubic worst case space complexity.<br>
 * (Remark: This could be done in a much more efficient way with only using two
 * pairs of indexes instead of copying all the elements of the clipping into a
 * new {@linkplain Contour} object. Feel free to improve this!)
 * 
 * @author Alexander Baumgartner
 */
public class Contour implements Cloneable {
	private ContourAtom[] elem = new ContourAtom[16];
	private int size;
	private Contour[][] clipCache = new Contour[16][];

	public void add(int i, int j, TermAtom tLeft, TermAtom tRight) {
		add(new ContourAtom(i, j, tLeft, tRight));
	}

	void add(ContourAtom ce) {
		if (size == elem.length)
			elem = Arrays.copyOf(elem, size + (size >> 1));
		elem[size] = ce;
		size++;
	}

	/**
	 * Computes a rectangular clip which contains all the elements of this
	 * contour inside the rectangle (0,0), ({@linkplain ContourAtom#i}-1,
	 * {@linkplain ContourAtom#j}-1). The computed clip will be cached, so that
	 * this computation is only performed once.<br>
	 * If iterAll is "false" then an admissibility test between the given
	 * {@linkplain ContourAtom} and all the elements in the clip will be
	 * performed and the clip may be empty if there is no admissible
	 * combination.
	 */
	public Contour clip(ContourAtom cElem) {
		int i = cElem.i - 1, j = cElem.j - 1;
		int oldLenI = clipCache.length;
		if (oldLenI <= i)
			clipCache = Arrays.copyOf(clipCache, i + 1);

		Contour[] clipCacheI = clipCache[i];
		if (clipCacheI == null)
			clipCacheI = clipCache[i] = new Contour[j + 1];
		else if (clipCacheI.length <= j)
			clipCacheI = clipCache[i] = Arrays.copyOf(clipCacheI, j + 1);

		if (clipCacheI[j] == null)
			computeClip(cElem, clipCacheI[j] = new Contour());
		return clipCacheI[j];
	}

	private Contour computeClip(ContourAtom cElem, Contour clip) {
		int i = cElem.i - 1, j = cElem.j - 1;
		TermAtom leftJ = cElem.atom.atomLeft;
		TermAtom rightJ = cElem.atom.atomRight;
		for (int idx = 0, n = size; idx < n; idx++) {
			ContourAtom ce = elem[idx];
			if (ce.i > i)
				return clip;
			if (ce.j <= j
					&& ce.atom.atomLeft.isAncestor(leftJ) == ce.atom.atomRight
							.isAncestor(rightJ))
				clip.add(ce);
		}
		return clip;
	}

	public int size() {
		return size;
	}

	public ContourAtom get(int i) {
		return elem[i];
	}

	public void clear() {
		size = 0;
		clipCache = new Contour[clipCache.length][];
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder(size * 8);
		for (int i = 0; i < size; i++)
			sb.append(elem[i]);
		return sb.toString();
	}

	public class ContourAtom implements Cloneable {
		protected int i, j;
		protected AlignmentAtom atom;

		ContourAtom(int i, int j, TermAtom left, TermAtom right) {
			this(i, j);
			atom = new AlignmentAtom(left, right);
		}

		ContourAtom(int i, int j) {
			this.i = i;
			this.j = j;
		}

		@Override
		public String toString() {
			return atom.atomLeft + "[" + i + "," + j + "]";
		}
	}
}