/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.data;

import java.util.Arrays;

import at.jku.risc.stout.urauc.data.Alignment.AlignmentAtom;
import at.jku.risc.stout.urauc.data.atom.TermAtom;
import at.jku.risc.stout.urauc.util.TinyList;

/**
 * An {@linkplain Alignment} is a list of {@linkplain AlignmentAtom}s.
 * 
 * @author Alexander Baumgartner
 */
public class Alignment extends TinyList<AlignmentAtom> {
	private AlignmentAtom[] atoms;

	/**
	 * Creates an alignment with initial capacity 16.
	 */
	public Alignment() {
		this(16);
	}

	/**
	 * Creates an alignment with the given initial capacity.
	 */
	public Alignment(int capacity) {
		this.atoms = new AlignmentAtom[capacity];
	}

	/**
	 * Sets the absolute size and increases the capacity if necessary.
	 */
	private void setSizeAbsolut(int size) {
		if (size > atoms.length) {
			atoms = Arrays.copyOfRange(atoms, start, size - start);
			size -= start;
			start = 0;
		}
		this.size = size;
	}

	/**
	 * Sets the relative size and increases the capacity if necessary.
	 */
	public void setSize(int size) {
		setSizeAbsolut(size + start);
	}

	@Override
	public Alignment clone() {
		int alignSize = size();
		Alignment clone = new Alignment(alignSize);
		clone.size = alignSize;
		for (int i = 0; i < alignSize; i++)
			clone.atoms[i] = get(i).clone();
		return clone;
	}

	@Override
	protected AlignmentAtom getAbsolut(int i) {
		return atoms[i];
	}

	/**
	 * Adds a new {@linkplain AlignmentAtom} to this {@linkplain Alignment}.
	 */
	public void addAtom(AlignmentAtom atom) {
		if (size == atoms.length) {
			atoms = Arrays.copyOfRange(atoms, start, size + (size >> 1));
			size -= start;
			start = 0;
		}
		atoms[size] = atom;
		size++;
	}

	/**
	 * Resets the alignment to the initial state.
	 */
	public void reset() {
		start = 0;
		size = 0;
	}

	public void incAllRootIdxLeft() {
		for (int i = start; i < size; i++)
			atoms[i].idxIncLeft();
	}

	public void incAllRootIdxRight() {
		for (int i = start; i < size; i++)
			atoms[i].idxIncRight();
	}

	public void addAllRootIdx(int addLeft, int addRight) {
		for (int i = start; i < size; i++) {
			AlignmentAtom atom = atoms[i];
			atom.idxLeftTopAdd(addLeft);
			atom.idxRightTopAdd(addRight);
		}
	}

	/**
	 * Sets the {@linkplain AlignmentAtom} at the given (relative) position if
	 * the atom is admissible with the elements behind it (the elements with
	 * higher index). Returns true if the admissibility test was positive.
	 */
	public boolean setAtomIfTailAdmissible(int idx, AlignmentAtom atom) {
		idx += start;
		if (isTailAdmissible(idx, atom.atomLeft, atom.atomRight)) {
			atoms[idx] = atom;
			return true;
		}
		return false;
	}

	/**
	 * Tests whether the alignment is admissible or not. This test has cubic
	 * complexity.
	 */
	public boolean isAdmissible() {
		for (int i = start, n1 = size - 1; i < n1; i++)
			if (!isTailAdmissible(i, atoms[i].atomLeft, atoms[i].atomRight))
				return false;
		return true;
	}

	/**
	 * Tests whether the given {@linkplain TermAtom} pair is admissible with the
	 * elements behind it (the elements with higher index). The index specifies
	 * the absolute position.
	 */
	protected boolean isTailAdmissible(int idx, TermAtom left, TermAtom right) {
		boolean isAncestor = true;
		for (int j = idx + 1; j < size; j++) {
			TermAtom leftJ = atoms[j].atomLeft;
			TermAtom rightJ = atoms[j].atomRight;
			// collision between two elements
			if (isAncestor) {
				isAncestor = left.isAncestor(leftJ);
				if (isAncestor != right.isAncestor(rightJ))
					return false;
			}
			// collision between three elements
			if (!isAncestor && isCollosion3Elem(j, left, right, leftJ, rightJ))
				return false;
		}
		return true;
	}

	/**
	 * Test whether there is a collision between three elements. Two pairs of
	 * elements are given and the third one is iterated between the given
	 * absolute index and {@linkplain #sizeAbsolut()}.
	 */
	protected boolean isCollosion3Elem(int idx, TermAtom leftI,
			TermAtom rightI, TermAtom leftJ, TermAtom rightJ) {
		int leftCommonIJ = leftI.commonAncestor(leftJ);
		int rightCommonIJ = rightI.commonAncestor(rightJ);
		if (leftCommonIJ == 0 && rightCommonIJ == 0)
			return false;
		for (int k = idx + 1; k < size; k++) {
			int leftCommonJK = leftJ.commonAncestor(atoms[k].atomLeft);
			int rightCommonJK = rightJ.commonAncestor(atoms[k].atomRight);
			if (leftCommonIJ > leftCommonJK && rightCommonIJ < rightCommonJK)
				return true;
			if (leftCommonIJ < leftCommonJK && rightCommonIJ > rightCommonJK)
				return true;
		}
		return false;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for (int i = start; i < size; i++)
			sb.append(' ').append(atoms[i].toString()); // GWT needs toString
		return sb.length() == 0 ? "" : sb.substring(1);
	}

	/**
	 * An {@linkplain AlignmentAtom} consists of two {@linkplain TermAtom}s. One
	 * atom is of the left hedge and the other one is of the right hedge.
	 * 
	 * @author Alexander Baumgartner
	 */
	public static class AlignmentAtom implements Cloneable {
		public TermAtom atomLeft, atomRight;
		private int leftIdxStart;
		private int rightIdxStart;
		private int leftIdxAdd;
		private int rightIdxAdd;

		public AlignmentAtom(TermAtom atomLeft, TermAtom atomRight) {
			this.atomLeft = atomLeft;
			this.atomRight = atomRight;
		}

		public void set(TermAtom atomLeft, TermAtom atomRight) {
			this.atomLeft = atomLeft;
			this.atomRight = atomRight;
			leftIdxStart = 0;
			rightIdxStart = 0;
			leftIdxAdd = 0;
			rightIdxAdd = 0;
		}

		public int getIdxLeftTop() {
			return atomLeft.getPositionIndex()[leftIdxStart] + leftIdxAdd;
		}

		public int getIdxRightTop() {
			return atomRight.getPositionIndex()[rightIdxStart] + rightIdxAdd;
		}

		public void idxLeftTopAdd(int toAdd) {
			leftIdxAdd += toAdd;
		}

		public void idxRightTopAdd(int toAdd) {
			rightIdxAdd += toAdd;
		}

		public void idxIncLeft() {
			leftIdxAdd = 0;
			leftIdxStart++;
		}

		public void idxIncRight() {
			rightIdxAdd = 0;
			rightIdxStart++;
		}

		public void idxInc() {
			idxIncLeft();
			idxIncRight();
		}

		public int getIdxLeftDepth() {
			return atomLeft.getPositionIndex().length - leftIdxStart;
		}

		public int getIdxRightDepth() {
			return atomRight.getPositionIndex().length - rightIdxStart;
		}

		@Override
		public AlignmentAtom clone() {
			try {
				AlignmentAtom clone = (AlignmentAtom) super.clone();
				clone.atomLeft = atomLeft.clone();
				clone.atomRight = atomRight.clone();
				return clone;
			} catch (CloneNotSupportedException e) {
				throw new RuntimeException(e);
			}
		}

		@Override
		public String toString() {
			return atomLeft.getName() + "<"
					+ atomLeft.getPositionString(leftIdxStart, leftIdxAdd)
					+ ", "
					+ atomRight.getPositionString(rightIdxStart, rightIdxAdd)
					+ ">";
		}
	}
}