/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.data;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.Collections;
import java.util.Set;

import at.jku.risc.stout.urauc.algo.AntiUnifyProblem;
import at.jku.risc.stout.urauc.algo.IllegalAlignmentException;
import at.jku.risc.stout.urauc.data.Alignment.AlignmentAtom;
import at.jku.risc.stout.urauc.data.atom.ContextVar;
import at.jku.risc.stout.urauc.data.atom.FunctionSymbol;
import at.jku.risc.stout.urauc.data.atom.FunctionVar;
import at.jku.risc.stout.urauc.data.atom.HedgeVar;
import at.jku.risc.stout.urauc.data.atom.Hole;
import at.jku.risc.stout.urauc.data.atom.TermAtom;
import at.jku.risc.stout.urauc.data.atom.TermVar;
import at.jku.risc.stout.urauc.util.DataStructureFactory;

/**
 * Tiny term parser with the following rules:
 * <ul>
 * <li>[A-Za-z1-9_.] are allowed for variable names and function names. Names
 * can be of any length.
 * <li>If the name is followed by an opening parenthesis a
 * {@linkplain FunctionSymbol} will be generated.
 * <ul>
 * <li>If the function name starts with an upper case letter and the first
 * letter of the name is within [{@linkplain #VARIABLE_START}, z], it will be
 * transformed into a {@linkplain ContextVar context variable}.
 * <li>Else if the function name starts with an upper case letter, it will be
 * transformed into a {@linkplain FunctionVar function variable}.
 * </ul>
 * Parameter inside the pair of parenthesis are wrapped in a {@linkplain Hedge}
 * <li>Else if the name starts with {@linkplain #HOLE_NAME} a {@linkplain Hole}
 * will be created.
 * <li>Else if the first letter of the name is within [
 * {@linkplain #VARIABLE_START}, z] an {@linkplain TermVar individual variable}
 * will be generated.
 * <ul>
 * <li>If the variable name starts with an upper case letter it will become a
 * {@linkplain HedgeVar hedge variable}. Otherwise an {@linkplain TermVar
 * individual variable} will be instantiates.
 * </ul>
 * <li>Else a {@linkplain FunctionSymbol} with zero parameters (=constant) will
 * be generated.
 * </ul>
 * 
 * @author Alexander Baumgartner
 */
public class InputParser<T extends Equation> {
	/**
	 * default = 'u' => A word starting with 'u-z' will become a variable.<br>
	 * {@linkplain InputParser#VARIABLE_START} has to be a lower case letter!
	 * Context variables are denoted by upper case letters.
	 */
	public static char VARIABLE_START = 'u';
	/**
	 * default = '@' => Any word starting with '@' will become a hole.
	 */
	public static char HOLE_NAME = '@';
	public static char OPENING_PARENTHESIS = '(';
	public static char CLOSING_PARENTHESIS = ')';
	private int codePoint;
	private EquationSystem<T> sys;
	private NodeFactory nodeFactory = new NodeFactory();

	public InputParser(EquationSystem<T> sys) {
		this.sys = sys;
	}

	/**
	 * Parses a system of anti-unification equations which are separated by
	 * semicolon. Use {@linkplain #parseHedgeEquation(Reader, Reader)} if
	 * possible! It performs much better and saves space by reading directly
	 * from the data streams!!!
	 */
	public void parseEquationSystem(String aupsIn)
			throws MalformedTermException {
		try {
			for (String aup : aupsIn.replace('[', '(').replace(']', ')')
					.split("\\s*;\\s*")) {
				String[] terms = aup
						.split("[^\\._a-zA-Z0-9\\(\\)]*=[^\\._a-zA-Z0-9\\(\\)]*");
				if (terms.length != 2)
					throw new MalformedTermException(
							"The input is not in equation form.");
				parseHedgeEquation(new StringReader(terms[0]),
						new StringReader(terms[1]));
			}
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * A hedge equation has the form:<br>
	 * (s1, s2,...) =^= (t1, t2,...)<br>
	 * 
	 * @param leftHedge
	 *            The left sequence (s1, s2,...)
	 * @param rightHedge
	 *            The right sequence (t1, t2,...)
	 */
	public T parseHedgeEquation(Reader leftHedge, Reader rightHedge)
			throws IOException, MalformedTermException {
		T aup = sys.newEquation();
		parseEquationSequence(leftHedge, aup, true);
		parseEquationSequence(rightHedge, aup, false);
		aup.init();
		sys.add(aup);
		reset();
		return aup;
	}

	private void parseEquationSequence(Reader in, T aup, boolean left)
			throws IOException, MalformedTermException {
		do {
			TermNode node = parseTerm(in);
			if (node != null) {
				if (left)
					aup.addLeft(node);
				else
					aup.addRight(node);
			}
		} while (codePoint != -1 && !isNameChar(codePoint));
	}

	/**
	 * Use {@linkplain #parseTerm(Reader)} if possible! It performs much better
	 * and saves space by reading directly from the data stream!!!
	 */
	public TermNode parseTerm(String in) throws IOException,
			MalformedTermException {
		return parseTerm(new StringReader(in));
	}

	public TermNode parseTerm(Reader in) throws IOException,
			MalformedTermException {
		if (nextNameChar(in)) // codePoint = "first letter"
			return parseIt(in);
		return null;
	}

	private TermNode parseIt(Reader in) throws IOException,
			MalformedTermException {
		String name = parseTermName(in).toString();
		if (name.length() == 0)
			throw new MalformedTermException(
					"Missing parenthesis at the input term");
		char nameLetter1 = name.charAt(0);
		if (codePoint == OPENING_PARENTHESIS) {
			nodeFactory.pushHedge();
			while (codePoint != CLOSING_PARENTHESIS
					&& next(in) != CLOSING_PARENTHESIS)
				nodeFactory.addToHedge(parseIt(in));
			Hedge hedge = nodeFactory.popHedge();
			next(in); // consume closing bracket
			if (Character.toLowerCase(nameLetter1) >= VARIABLE_START)
				return nodeFactory.createContextVar(name, hedge);
			return nodeFactory.createFunction(name, hedge);
		} else if (nameLetter1 == HOLE_NAME) {
			return nodeFactory.createHole(name);
		} else if (Character.toLowerCase(nameLetter1) >= VARIABLE_START) {
			return nodeFactory.createHedgeVar(name);
		}
		return nodeFactory.createConstant(name);
	}

	private boolean nextNameChar(Reader in) throws IOException {
		while (!isNameChar(next(in)))
			if (codePoint == -1)
				return false;
		return true;
	}

	private StringBuilder parseTermName(Reader in) throws IOException {
		StringBuilder name = new StringBuilder();
		for (; isNameChar(codePoint); next(in))
			name.appendCodePoint(codePoint);
		return name;
	}

	private int next(Reader in) throws IOException {
		codePoint = in.read();
		return Character.isWhitespace(codePoint) ? next(in) : codePoint;
	}

	public boolean isNameChar(int codePoint) {
		return codePoint == '_' || codePoint == '.' || codePoint == HOLE_NAME
				|| Character.isLetterOrDigit(codePoint);
	}

	public static Set<TermAtom> parseAtomSet(String atomsCSV) {
		if (atomsCSV == null || atomsCSV.trim().length() == 0)
			return Collections.emptySet();
		Set<TermAtom> retSet = DataStructureFactory.$.newSet();
		for (String symb : atomsCSV.trim().split("\\W+"))
			retSet.add(new FunctionSymbol(symb));
		return retSet;
	}

	public static Alignment parseAlignment(AntiUnifyProblem aup,
			String alignment, boolean checkAdmissible)
			throws IllegalAlignmentException {
		Alignment al = new Alignment();
		if (alignment == null || alignment.trim().length() == 0)
			return al;
		if (aup == null)
			throw new IllegalAlignmentException(
					"The corresponding AUP must not be null");
		for (String elem : alignment.split("[>]")) {
			int iOf1 = elem.indexOf('<');
			int iOf2 = elem.indexOf(',', iOf1);
			if (iOf2 == -1)
				iOf2 = elem.indexOf(';', iOf1);
			String name = elem.substring(0, iOf1).trim().intern();
			String[] intS = elem.substring(iOf1 + 1, iOf2).split("\\.");
			int[] intL = new int[intS.length];
			for (int i = intS.length - 1; i >= 0; i--)
				intL[i] = Integer.parseInt(intS[i].trim()) - 1;
			intS = elem.substring(iOf2 + 1).split("\\.");
			int[] intR = new int[intS.length];
			for (int i = intS.length - 1; i >= 0; i--)
				intR[i] = Integer.parseInt(intS[i].trim()) - 1;
			TermAtom left = aup.getLeft().get(intL, 0);
			TermAtom right = aup.getRight().get(intR, 0);
			if (left == null || right == null)
				throw new IllegalAlignmentException(
						"Illegal position at alignment element "
								+ (al.size() + 1));
			if (left.getName() != name || right.getName() != name)
				throw new IllegalAlignmentException(
						"Different symbols at alignment element "
								+ (al.size() + 1));
			al.addAtom(new AlignmentAtom(left, right));
		}
		if (checkAdmissible && !al.isAdmissible())
			throw new IllegalAlignmentException("Alignment is not admissible");
		return al;
	}

	/**
	 * Resets the state of the parser.
	 */
	public void reset() {
		codePoint = 0;
	}
}
