/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.data;

import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;
import java.util.Map;
import java.util.Set;

import at.jku.risc.stout.urauc.data.atom.Hole;
import at.jku.risc.stout.urauc.data.atom.TermAtom;
import at.jku.risc.stout.urauc.data.atom.Variable;
import at.jku.risc.stout.urauc.util.Printable;

/**
 * A term node consists of a {@linkplain TermAtom} and a {@linkplain Hedge}.
 * 
 * @author Alexander Baumgartner
 */
public class TermNode extends Printable implements Cloneable,
		Comparable<TermNode> {
	boolean ignore;
	private TermAtom atom;
	private Hedge hedge;
	public static final int[] POSITION_E = new int[] {};

	/**
	 * Instantiates a new term node. null values are transformed to unique
	 * objects which represent the null value. Such that no
	 * {@linkplain NullPointerException} can occur.
	 * 
	 * @see TermAtom#nullAtom
	 * @see Hedge#nullHedge
	 */
	public TermNode(TermAtom atom, Hedge hedge) {
		setAtom(atom);
		setHedge(hedge);
	}

	/**
	 * Instantiates a new term node with an empty hedge.
	 * 
	 * @see TermNode#TermNode(TermAtom, Hedge)
	 */
	public TermNode(TermAtom atom) {
		setAtom(atom);
		setHedge(null);
	}

	public TermAtom getAtom() {
		return atom;
	}

	/**
	 * Null values are transformed to unique objects which represent the null
	 * value. So that no {@linkplain NullPointerException} can occur.
	 * 
	 * @see TermAtom#nullAtom
	 */
	public void setAtom(TermAtom atom) {
		this.atom = atom == null ? TermAtom.nullAtom : atom;
	}

	public Hedge getHedge() {
		return hedge;
	}

	/**
	 * Null values are transformed to unique objects which represent the null
	 * value. So that no {@linkplain NullPointerException} can occur.
	 * 
	 * @see Hedge#nullHedge
	 */
	public void setHedge(Hedge hedge) {
		this.hedge = hedge == null ? Hedge.nullHedge : hedge;
	}

	/**
	 * Tests whether the atom of this term node is the null atom.
	 */
	public boolean isNullAtom() {
		return atom == TermAtom.nullAtom;
	}

	/**
	 * Tests whether the atom of this term node is the hole.
	 */
	public boolean isHole() {
		return atom instanceof Hole;
	}

	@Override
	public boolean equals(Object other) {
		if (other instanceof TermNode) {
			TermNode oNode = (TermNode) other;
			return atom.equals(oNode.atom) && hedge.equals(oNode.hedge);
		}
		return false;
	}

	@Override
	public int hashCode() {
		return atom.hashCode() + 31 * hedge.hashCode();
	}

	@Override
	public TermNode clone() {
		return new TermNode(atom.clone(), hedge.clone());
	}

	/**
	 * Substitutes one {@linkplain Variable} by a {@linkplain TermNode}
	 * recursively and returns the new {@linkplain TermNode}.
	 */
	public TermNode substitute(Variable from, TermNode to) {
		for (int i = hedge.size() - 1; i >= 0; i--)
			hedge.replace(i, hedge.get(i).substitute(from, to));
		return atom.substitute(from, to, this);
	}

	/**
	 * Applies a substitution recursively and returns the new
	 * {@linkplain TermNode}.
	 */
	public TermNode apply(Map<Variable, TermNode> sigma) {
		for (int i = hedge.size() - 1; i >= 0; i--)
			hedge.replace(i, hedge.get(i).apply(sigma));
		return atom.apply(sigma, this);
	}

	/**
	 * Replaces the hole in this context with the given {@linkplain TermNode}
	 * (term | hedge | context). Returns null if no hole has been found,
	 * otherwise the new {@linkplain TermNode} will be returned.
	 */
	public TermNode replaceHole(TermNode plugIn) {
		if (atom instanceof Hole)
			return plugIn.clone();

		for (int i = hedge.size() - 1; i >= 0; i--) {
			TermNode replaced = hedge.get(i).replaceHole(plugIn);
			if (replaced != null) {
				hedge.replace(i, replaced);
				if (atom.isCommutative())
					hedge.setCommutative(true);
				return this;
			}
		}
		return null;
	}

	/**
	 * Initializes all the position indexes recursively starting with the given
	 * position parentIndex.childPosition.
	 */
	public void initPosition(int[] parentIndex, int pos) {
		if (!isNullAtom()) {
			int len = parentIndex.length;
			parentIndex = Arrays.copyOf(parentIndex, len + 1);
			parentIndex[len] = pos;
			atom.setPositionIndex(parentIndex);
		}
		hedge.initPosition(parentIndex);
	}

	public void setPositionRecurse(int idx, int value) {
		atom.getPositionIndex()[idx] = value;
		for (int i = hedge.size(); --i >= 0;)
			hedge.get(i).setPositionRecurse(idx, value);
	}

	public void sortCommutative(boolean sortAll) {
		hedge.sortCommutative(sortAll);
	}

	public void makeCommutative(Set<TermAtom> commutativeSymbols) {
		atom.setCommutative(commutativeSymbols.contains(atom));
		hedge.makeCommutative(commutativeSymbols, atom.isCommutative());
	}

	public boolean nextCommutativeArrangement() {
		return hedge.nextCommutativeArrangement();
	}

	@Override
	public void print(Writer out) throws IOException {
		atom.print(out);
		hedge.print(out);
	}

	/**
	 * Traverses the term tree and executes the callback function at every node.
	 * 
	 * @see TraverseCallBack
	 */
	public boolean traverse(TraverseCallBack callBack) {
		if (callBack.exec(atom, hedge))
			return true;
		for (int i = 0, n = hedge.size(); i < n; i++)
			if (hedge.get(i).traverse(callBack))
				return true;
		return false;
	}

	/**
	 * This class may be used to
	 * {@linkplain TermNode#traverse(TraverseCallBack) traverse} a term tree and
	 * execute an arbitrary operation on every node.
	 * 
	 * @author Alexander Baumgartner
	 */
	public abstract class TraverseCallBack {
		/**
		 * Propagation will stop as soon as this method returns true.
		 */
		public abstract boolean exec(TermAtom atom, Hedge hedge);
	}

	public TermAtomList word() {
		TermAtomList atomList = new TermAtomList();
		word(atomList);
		return atomList;
	}

	public void word(TermAtomList atomList) {
		if (!isNullAtom())
			atomList.add(atom);
		hedge.word(atomList);
	}

	public TermAtom get(int[] position, int idx) {
		return idx == position.length ? atom : hedge.get(position, idx);
	}

	/**
	 * This compare method treads commutativity by ordered argument sequences.
	 */
	public int compareTo(TermNode other) {
		int ret = atom.compareTo(other.atom);
		if (ret != 0)
			return ret;
		return hedge.compareTo(other.hedge);
	}

	public void normalizeVarNames(Map<String, String> renaming) {
		atom.normalizeVarNames(renaming);
		hedge.normalizeVarNames(renaming);
	}

	public void makeAssociative(Set<TermAtom> associativeSymbols) {
		getHedge().makeAssociative(associativeSymbols,
				associativeSymbols.contains(atom) ? atom : null);
	}
}
